!--------1---------2---------3---------4---------5---------6---------7---------8
!
!  Module   at_rpath_mod
!> @brief   String method using replicas
!! @authors Yasuaki Komuro (YK), Takaharu Mori (TM), Norio Takase (NT)
! 
!  (c) Copyright 2014 RIKEN. All rights reserved.
!
!--------1---------2---------3---------4---------5---------6---------7---------8

#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif

module at_rpath_mod

  use at_energy_mod
  use at_energy_str_mod
  use at_energy_restraints_mod
  use at_md_leapfrog_mod
  use at_md_vverlet_mod
  use at_dynamics_str_mod
  use at_dynvars_str_mod
  use at_dynvars_mod
  use at_ensemble_str_mod
  use at_rpath_str_mod
  use at_restraints_str_mod
  use at_constraints_str_mod
  use at_boundary_str_mod
  use at_boundary_mod
  use at_pairlist_str_mod
  use at_pairlist_mod
  use at_enefunc_str_mod
  use at_enefunc_restraints_mod
  use dihedral_libs_mod
  use timers_mod
  use at_output_str_mod
  use at_output_mod
  use molecules_str_mod
  use fileio_rst_mod
  use fileio_top_mod
  use fileio_par_mod
  use fileio_gpr_mod
  use fileio_pdb_mod
  use fileio_psf_mod
  use fileio_crd_mod
  use fileio_mode_mod
  use fileio_control_mod
  use select_mod
  use select_atoms_mod
  use select_atoms_str_mod
  use string_mod
  use random_mod
  use messages_mod
  use mpi_parallel_mod
  use constants_mod
#ifdef MPI
  use mpi
#endif

  implicit none

  private
  ! structures
  type, public :: s_rpath_info
    integer                           :: dimension    = 0
    integer                           :: rpath_period = 0
    integer                           :: nreplica     = 0
    real(wp)                          :: delta        = 0.0_wp
    real(wp)                          :: smooth       = 0.0_wp
    logical                           :: fix_terminal = .false.
    logical                           :: use_restart  = .true.
    integer,              allocatable :: rest_function(:)
  end type s_rpath_info

  ! subroutines
  public  :: show_ctrl_rpath
  public  :: read_ctrl_rpath
  public  :: define_nreplica
  public  :: setup_rpath
  public  :: run_rpath

  ! private subroutines
  private :: assign_condition
  private :: evolve
  private :: reparametrize

contains

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    show_ctrl_rpath
  !> @brief        show RPATH section usage
  !! @authors      NT, TM, YK
  !! @param[in]    show_all : show all usage or not
  !! @param[in]    run_mode : run mode string : "md", "min", "rpath"
  !
  !======1=========2=========3=========4=========5=========6=========7=========8
  
  subroutine show_ctrl_rpath(show_all, run_mode)

    ! formal arguments
    logical,                 intent(in)    :: show_all
    character(*),            intent(in)    :: run_mode

    if (show_all) then

      select case (run_mode)

      case ('rpath')

        write(MsgOut,'(A)') '[RPATH]'
        write(MsgOut,'(A)') 'nreplica          = 4'
        write(MsgOut,'(A)') 'rpath_period      = 1000'
        write(MsgOut,'(A)') 'delta             = 0.05'
        write(MsgOut,'(A)') 'smooth            = 0.1'
        write(MsgOut,'(A)') 'fix_terminal      = NO'
        write(MsgOut,'(A)') 'rest_function     = 1 2'
        write(MsgOut,'(A)') 'use_restart       = YES'
        write(MsgOut,'(A)') ''

      end select

    else

      select case (run_mode)

      case ('rpath')

        write(MsgOut,'(A)') '[RPATH]'
        write(MsgOut,'(A)') 'nreplica          = 4'
        write(MsgOut,'(A)') 'rpath_period      = 1000'
        write(MsgOut,'(A)') 'delta             = 0.05'
        write(MsgOut,'(A)') 'smooth            = 0.1'
        write(MsgOut,'(A)') 'fix_terminal      = NO'
        write(MsgOut,'(A)') 'rest_function     = 1 2'
        write(MsgOut,'(A)') 'use_restart       = YES'
        write(MsgOut,'(A)') ''

      end select

    end if

    return

  end subroutine show_ctrl_rpath
  
  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    read_ctrl_rpath
  !> @brief        read RPATH section in the control file
  !! @authors      TM, YK
  !! @param[in]    handle     : unit number of control file
  !! @param[out]   rpath_info : RPATH section control parameters information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine read_ctrl_rpath(handle, rpath_info)

    ! parameters
    character(*),            parameter     :: Section = 'Rpath'

    ! formal arguments
    integer,                 intent(in)    :: handle
    type(s_rpath_info),      intent(inout) :: rpath_info

    ! local variables
    integer                  :: i
    character(30)            :: cdim, partmp, vartmp, fnctmp
    character(MaxLine)       :: rest_function_char

    ! read parameters from control file
    !
    call begin_ctrlfile_section(handle, Section)

    call read_ctrlfile_integer(handle, Section, 'nreplica',       &
                               rpath_info%nreplica)
    call read_ctrlfile_integer(handle, Section, 'rpath_period', &
                               rpath_info%rpath_period)
    call read_ctrlfile_real(handle, Section, 'delta',  &
                            rpath_info%delta)
    call read_ctrlfile_real(handle, Section, 'smooth', &
                            rpath_info%smooth)
    call read_ctrlfile_logical(handle, Section, 'fix_terminal',  &
                            rpath_info%fix_terminal)
    call read_ctrlfile_string (handle, Section, 'rest_function', &
                               rest_function_char)
    call read_ctrlfile_logical(handle, Section, 'use_restart',  &
                            rpath_info%use_restart)

    call end_ctrlfile_section(handle)

    ! error check
    !
    if (rpath_info%nreplica == 0) &
      call error_msg('Read_Ctrl_Rpath> nreplica should be > 0 in [RPATH]')

    ! get dimension from the number of numerics in rest_function_char
    !
    rpath_info%dimension = split_num(rest_function_char)

    ! read rest_function
    !
    allocate(rpath_info%rest_function(rpath_info%dimension))
    rpath_info%rest_function(1:rpath_info%dimension) = 0
    call split(rpath_info%dimension, rpath_info%dimension, rest_function_char, rpath_info%rest_function)

    ! write parameters to MsgOut
    !
    if (main_rank) then
      write(MsgOut,'(A)') 'Read_Ctrl_Rpath> Rpath information'
      write(MsgOut,'(A20,I10)') &
        '  dimension       = ', rpath_info%dimension
      write(MsgOut,'(A20,I10)') &
        '  nreplica        = ', rpath_info%nreplica
      write(MsgOut,'(A20,I10)') &
        '  rpath_period    = ', rpath_info%rpath_period
      write(MsgOut,'(A20,F10.5)') &
        '  delta           = ', rpath_info%delta
      if (rpath_info%fix_terminal) then
        write(MsgOut,'(A)') '  fix_terminal    = YES'
      else
        write(MsgOut,'(A)') '  fix_terminal    = NO'
      end if
      if (rpath_info%use_restart) then
        write(MsgOut,'(A)') '  use_restart     = YES'
      else
        write(MsgOut,'(A)') '  use_restart     = NO'
      end if
      write(MsgOut,'(A20,F10.3)') &
        '  smooth          = ', rpath_info%smooth
      write(MsgOut,'(A20,A)') &
        '  rest_function   = ', trim(rest_function_char)
      write(MsgOut,'(A)') ''

    end if

    return

  end subroutine read_ctrl_rpath

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    define_nreplica
  !> @brief        setup structure of replica
  !! @authors      TM, YK
  !! @param[in]    rpath_info : RPATH section control parameters information
  !! @param[out]   replica  : replica information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine define_nreplica(rpath_info, rpath)

    ! formal arguments
    type(s_rpath_info),       intent(in)    :: rpath_info
    type(s_rpath),            intent(inout) :: rpath

    rpath%nreplica = rpath_info%nreplica

    return

  end subroutine define_nreplica

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    setup_rpath
  !> @brief        setup RPATH
  !! @authors      TM, YK
  !! @param[in]    rpath_info : RPATH section control parameters information
  !! @param[in]    rst        : rst data
  !! @param[in]    dynamics   : dynamics information
  !! @param[in]    molecule   : molecule information
  !! @param[inout] restraints : restraints information
  !! @param[inout] enefunc    : potential energy functions information
  !! @param[inout] rpath      : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine setup_rpath(rpath_info, rst, dynamics, molecule, &
                         restraints, enefunc, rpath)

    ! formal arguments
    type(s_rpath_info),      intent(in)    :: rpath_info
    type(s_rst),     target, intent(in)    :: rst
    type(s_dynamics),target, intent(in)    :: dynamics
    type(s_molecule),target, intent(in)    :: molecule
    type(s_restraints),      intent(inout) :: restraints
    type(s_enefunc), target, intent(inout) :: enefunc
    type(s_rpath),   target, intent(inout) :: rpath

    ! local variables
    integer                    :: i, j, k, m, dimno
    integer                    :: max_nreplica
    integer                    :: natmgrp
    character(2000)            :: param
    real(wp),      allocatable :: ddata(:)
    integer,       pointer     :: grouplist(:,:), numatoms(:), atomlist(:,:)
    real(wp),      pointer     :: mass(:)
    integer                    :: iatm, iatm_xyz
    integer                    :: replicaid
    integer                    :: ifunc, igroup

    mass          => molecule%mass
    grouplist     => enefunc%restraint_grouplist
    numatoms      => enefunc%restraint_numatoms
    atomlist      => enefunc%restraint_atomlist

    ! setup reparameterization period
    !
    if (rpath_info%rpath_period > 0) then
      if (mod(dynamics%nsteps, rpath_info%rpath_period) /= 0) then
        call error_msg('Setup_Rpath> mod(nsteps,rpath_period) must be zero')
      else
        rpath%equilibration_only = .false.
        rpath%rpath_period = rpath_info%rpath_period
        rpath%ncycle = dynamics%nsteps/rpath%rpath_period
      end if
    else if (rpath_info%rpath_period == 0) then
      rpath%equilibration_only = .true.
      rpath%rpath_period = dynamics%nsteps
      rpath%ncycle = 1
    else
      call error_msg('Setup_Rpath> rpath_period should be non-negative integer')
    end if

    ! setup basic parameters
    !
    rpath%nreplica     = rpath_info%nreplica
    max_nreplica       = rpath_info%nreplica
    rpath%delta        = rpath_info%delta
    rpath%smooth       = rpath_info%smooth
    rpath%fix_terminal = rpath_info%fix_terminal
    rpath%use_restart  = rpath_info%use_restart

    ! setup force constants and references
    !
    ifunc = rpath_info%rest_function(1)
    igroup = grouplist(1, ifunc)
    if(enefunc%restraint_kind(ifunc) == RestraintsFuncPOSI) then

      rpath%dimension = numatoms(igroup) * 3

      call alloc_rpath(rpath, RpathReplicas, rpath%dimension,    &
                    rpath%nreplica, max_nreplica)

      rpath%rest_function(1) = rpath_info%rest_function(1)
      rpath%rest_function(2:rpath%dimension) = 0

      call alloc_rpath(rpath, RpathUmbrellas, &
                    rpath%dimension, rpath%nreplica)

      replicaid = my_country_no + 1
      do i = 1, rpath%dimension
        iatm = atomlist(int((i-1)/3)+1,igroup)
        iatm_xyz = mod(i-1, 3) + 1
        do j = 1, rpath%nreplica
          rpath%rest_constants(1:4, i, replicaid) = enefunc%restraint_const_replica(j, ifunc)
          rpath%rest_reference(1:2, i, replicaid) = enefunc%restraint_refcoord(iatm_xyz, iatm)
        end do
      end do

    else

      rpath%dimension = rpath_info%dimension

      call alloc_rpath(rpath, RpathReplicas, rpath%dimension, &
                       rpath%nreplica, max_nreplica)

      rpath%rest_function(:) = rpath_info%rest_function(:)

      ! split rpath_info%parameters (char) into parameters (real or integer)
      !
      call alloc_rpath(rpath, RpathUmbrellas, &
                       rpath%dimension, rpath%nreplica)

      if (rpath%use_restart .and. allocated(rst%rest_reference)) then
        if(replica_main_rank .and. (size(rst%rest_reference(1,:,1))==rpath%dimension)) then
          call mpi_allgather(rst%rest_reference(1, :, my_country_no+1), rpath%dimension, mpi_wp_real,&
                             rpath%rest_reference(1, :, :),  rpath%dimension, mpi_wp_real,&
                             mpi_comm_airplane, ierror)
        end if
        call mpi_bcast(rpath%rest_reference(1, :, :), rpath%dimension*rpath%nreplica,&
                       mpi_wp_real, 0, mpi_comm_country, ierror)
        do i = 1, rpath%dimension
          ifunc = rpath_info%rest_function(i)
          do j = 1, rpath%nreplica
            rpath%rest_constants(1:4, i, j) = enefunc%restraint_const_replica(j, ifunc)
            rpath%rest_reference(2, i, j) = rpath%rest_reference(1, i, j)
            enefunc%restraint_ref_replica(j, ifunc) = rpath%rest_reference(1, i, j)
          end do
        end do
      else
        do i = 1, rpath%dimension
          ifunc = rpath_info%rest_function(i)
          do j = 1, rpath%nreplica
            rpath%rest_constants(1:4, i, j) = enefunc%restraint_const_replica(j, ifunc)
            rpath%rest_reference(1:2, i, j) = enefunc%restraint_ref_replica(j, ifunc)
          end do
        end do
      endif

    end if

    enefunc%restraint_replica_index(:) = 0
    do i = 1, rpath%dimension
      ifunc = rpath_info%rest_function(i)
      enefunc%restraint_replica_index(ifunc) = i
    end do

    ! assign restraint parameters for enefunc
    !
    call assign_condition(rpath, enefunc)

    ! determine the number of atoms involved in collective variable
    !
    ifunc = rpath_info%rest_function(1)
    igroup = grouplist(1, ifunc)
    select case(enefunc%restraint_kind(ifunc))

    case (RestraintsFuncPOSI)

      natmgrp = numatoms(igroup)

    case (RestraintsFuncPC:RestraintsFuncPCCOM)

      natmgrp = numatoms(igroup)

    case default

      natmgrp = enefunc%max_restraint_numgrps

    end select

    ! setup statistical variables in dynvars
    !
    enefunc%rpath_flag      = .true.
    enefunc%stats_count     = 0
    enefunc%stats_natom     = natmgrp
    enefunc%stats_dimension = rpath%dimension

    allocate(enefunc%stats_delta(1:rpath%dimension), &
             enefunc%stats_grad(1:3, 1:natmgrp, 1:rpath%dimension), &
             enefunc%stats_force(1:rpath%dimension), &
             enefunc%stats_metric(1:rpath%dimension,1:rpath%dimension), &
             enefunc%stats_atom(1:natmgrp, 1:rpath%dimension), &
             enefunc%stats_mass(1:natmgrp, 1:rpath%dimension))

    enefunc%stats_delta(:)    = 0.0_wp
    enefunc%stats_grad(:,:,:) = 0.0_wp
    enefunc%stats_force(:)    = 0.0_wp
    enefunc%stats_metric(:,:) = 0.0_wp
    enefunc%stats_atom(:,:)   = 0
    enefunc%stats_mass(:,:)   = 0.0_wp

    ! setup atom index and mass
    !
    ifunc = rpath_info%rest_function(1)
    igroup = grouplist(1, ifunc)
    select case(enefunc%restraint_kind(ifunc))

    case (RestraintsFuncPOSI)

      do i = 1, natmgrp
        j = atomlist(i, 1)
        ! j = atomlist(i,igroup)
        enefunc%stats_atom(i,:) = j
        enefunc%stats_mass(i,:) = mass(j)
      end do

    case (RestraintsFuncPC:RestraintsFuncPCCOM)

      do i = 1, natmgrp
        j = atomlist(i, 1)
        ! j = atomlist(i,igroup)
        enefunc%stats_atom(i,:) = j
        enefunc%stats_mass(i,:) = mass(j)
      end do

    case default

      do dimno = 1, rpath%dimension
        !igroup = grouplist(1, dimno)
        do i = 1, natmgrp
          !j = atomlist(i,igroup)
          j = atomlist(1,grouplist(i,dimno))
          enefunc%stats_atom(i,dimno) = j
          enefunc%stats_mass(i,dimno) = mass(j)
        end do
      end do

    end select

    ! write the summary of setup
    !
    if (main_rank) then

      write(MsgOut,'(A)') 'Setup_Rpath> Rpath information'
      write(MsgOut,'(A)') ''
      write(MsgOut,'(A)') '  Restraints'

      if (rpath%use_restart) then
        if (enefunc%restraint_kind(ifunc) == RestraintsFuncPOSI .or.  &
            (.not. allocated(rst%rest_reference))) then
          write(MsgOut,'(A)') '  use_restart flag is ignored.'
        else
          write(MsgOut,'(A)') '  References are replaced.'
          write(MsgOut, *)
          do i = 1, rpath%dimension
            write(MsgOut,'(a,i0,a,$)')'  constant', i, ' = '
            do j = 1, rpath%nreplica
              write(MsgOut, '(F10.4,$)') rpath%rest_constants(1, i, j)
            end do
            write(MsgOut, *)
            write(MsgOut,'(a,i0,a,$)')'  reference', i, ' = '
            do j = 1, rpath%nreplica
              write(MsgOut, '(F10.4,$)') rpath%rest_reference(1, i, j)
            end do
            write(MsgOut,'(A)') ''
            write(MsgOut,'(A)') ''
          end do
        endif
      endif

      ! do i = 1, rpath%dimension
      !   do j = 1, rpath%nreplica
      !     write(MsgOut,'(A,I4,A,I4,A,1F8.3,A,1F8.3)')     &
      !       '    REPLICA =', j, '  DIM = ', i, &
      !       '  Const = ',rpath%rest_constants(1, i, j), &
      !       '  Ref   = ',rpath%rest_reference(1, i, j)
      !   end do
      ! end do

      write(MsgOut,'(A)') ''
    end if

    return

  end subroutine setup_rpath

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    assign_condition
  !> @brief        control replica exchange
  !! @authors      TM, YK
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] rpath       : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine assign_condition(rpath, enefunc)

    ! formal arguments
    type(s_rpath),           intent(in)    :: rpath
    type(s_enefunc),         intent(inout) :: enefunc

    ! local variables
    integer                  :: i
    integer                  :: replicaid
    integer                  :: ifunc, igroup
    integer                  :: iatm, iatm_xyz

    replicaid = my_country_no + 1
    ifunc = rpath%rest_function(1)
    igroup = enefunc%restraint_grouplist(1, ifunc)
    if(enefunc%restraint_kind(ifunc) == RestraintsFuncPOSI) then

      do i = 1, rpath%dimension
        iatm = enefunc%restraint_atomlist(int((i-1)/3)+1,igroup)
        iatm_xyz = mod(i-1, 3) + 1
        enefunc%restraint_refcoord(iatm_xyz, iatm) = rpath%rest_reference(1, i, replicaid)
      end do

    else

      do i = 1, rpath%dimension
        ifunc = rpath%rest_function(i)
        enefunc%restraint_const(1:4, ifunc) = rpath%rest_constants(1:4, i, replicaid)
        enefunc%restraint_ref  (1:2, ifunc) = rpath%rest_reference(1:2, i, replicaid)
      end do

    end if

    return

  end subroutine assign_condition

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    run_rpath
  !> @brief        run string method
  !! @authors      TM, YK
  !! @param[inout] output      : output information
  !! @param[inout] molecule    : molecule information
  !! @param[inout] enefunc     : potential energy functions information
  !! @param[inout] dynvars     : dynamics variables information
  !! @param[inout] dynamics    : dynamics information
  !! @param[inout] pairlist    : pairlist information
  !! @param[inout] boundary    : boundary conditions information
  !! @param[inout] constraints : constraints information
  !! @param[inout] ensemble    : ensemble information
  !! @param[inout] rpath       : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine run_rpath(output, molecule, enefunc, dynvars, dynamics,    &
                      pairlist, boundary, constraints, ensemble, rpath)

    ! formal arguments
    type(s_output),           intent(inout) :: output
    type(s_molecule), target, intent(inout) :: molecule
    type(s_enefunc),          intent(inout) :: enefunc
    type(s_dynvars),  target, intent(inout) :: dynvars
    type(s_dynamics),         intent(inout) :: dynamics
    type(s_pairlist),         intent(inout) :: pairlist
    type(s_boundary),         intent(inout) :: boundary
    type(s_constraints),      intent(inout) :: constraints
    type(s_ensemble),         intent(inout) :: ensemble
    type(s_rpath),            intent(inout) :: rpath

    ! local variables
    integer                   :: i, j, k
    integer                   :: iloop_start, iloop_end
    integer                   :: replicaid

    ! Open output files
    !
    call open_output(output)
    DynvarsOut = output%logunit

    ! assign conditions to enefunc
    !
    call assign_condition(rpath, enefunc)

    ! update boundary and pairlist
    !
    call update_boundary(enefunc%table%table, &
      enefunc%pairlistdist, boundary)
    if (real_calc) then
      pairlist%allocation = .true.
      call update_pairlist(enefunc, boundary, dynvars%coord, pairlist)
    end if

    ! run rpath
    !
    do i = 1, rpath%ncycle
      dynamics%istart_step  = (i-1)*rpath%rpath_period + 1
      dynamics%iend_step    =  i   *rpath%rpath_period
      dynamics%initial_time = dynvars%time

      ! MD main loop
      !
      if (dynamics%integrator == IntegratorLEAP) then
        call leapfrog_dynamics(output, molecule, enefunc, dynvars, dynamics, &
                               pairlist, boundary, constraints, ensemble)
      else if (dynamics%integrator == IntegratorVVER) then
        call vverlet_dynamics (output, molecule, enefunc, dynvars, dynamics, &
                               pairlist, boundary, constraints, ensemble)
      end if

      ! path evolution
      !
      if (.not. rpath%equilibration_only) then
        call evolve(enefunc, rpath)
        call reparametrize(rpath)
        call assign_condition(rpath, enefunc)
      end if

      ! output dynamical variables
      !
      call output_rpath(output, molecule, dynamics, dynvars, boundary, rpath)
    end do
     
    ! close output files
    !
    call close_output(output)

    return

  end subroutine run_rpath

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    evolve
  !> @brief        evolve path
  !! @authors      YK, YM
  !! @param[inout] enefunc  : potential energy functions information
  !! @param[inout] rpath    : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine evolve(enefunc, rpath)

    ! formal arguments
    type(s_enefunc),  target, intent(inout) :: enefunc
    type(s_rpath),    target, intent(inout) :: rpath

    ! local variables
    integer                  :: dimno, dimno_i, dimno_j
    integer                  :: repid

    integer,         pointer :: count
    real(wp),        pointer :: force(:), metric(:,:)

    if(.not. replica_main_rank) return

    count  => enefunc%stats_count
    force  => enefunc%stats_force
    metric => enefunc%stats_metric

    repid = my_country_no + 1

    do dimno = 1, rpath%dimension
      force(dimno) = force(dimno) / real(count,wp)
    end do

    do dimno_i = 1, rpath%dimension
      do dimno_j = 1, rpath%dimension
        metric(dimno_i,dimno_j) = metric(dimno_i,dimno_j) / real(count,wp)
      end do
    end do

    force = matmul(metric, force)

    if(rpath%fix_terminal) then
      if((repid == 1) .or. (repid == rpath%nreplica)) then
        force(:) = 0.0_wp
      end if
    end if
    
    do dimno_i = 1, rpath%dimension
      rpath%rest_reference(1:2, dimno_i, repid) = &
        rpath%rest_reference(1:2, dimno_i, repid) + rpath%delta * force(dimno_i)
    end do

    count       = 0
    force(:)    = 0.0_wp
    metric(:,:) = 0.0_wp

    return

  end subroutine evolve

  !======1=========2=========3=========4=========5=========6=========7=========8
  !
  !  Subroutine    reparametrize
  !> @brief        reparametrize path
  !! @authors      YK, YM
  !! @param[inout] rpath    : RPATH information
  !
  !======1=========2=========3=========4=========5=========6=========7=========8

  subroutine reparametrize(rpath)

    ! formal arguments
    type(s_rpath),    target, intent(inout) :: rpath

    ! local variables
    integer                  :: i, j, k
    integer                  :: dimno, dimno_i, dimno_j
    integer                  :: repid, repid_i, repid_j

    real(dp),    allocatable :: path(:,:), path_smooth(:,:), path_reparm(:,:)
    real(dp),    allocatable :: path_leng(:), path_equi(:)

    real(dp),        pointer :: force(:), metric(:,:)
    real(dp),        pointer :: before_gather(:), after_gather(:)

    force         => rpath%force
    metric        => rpath%metric
    before_gather => rpath%before_gather
    after_gather  => rpath%after_gather

    repid = my_country_no + 1

    allocate(path(rpath%nreplica, rpath%dimension),&
             path_smooth(rpath%nreplica, rpath%dimension),&
             path_reparm(rpath%nreplica, rpath%dimension),&
             path_leng(rpath%nreplica), path_equi(rpath%nreplica))

    do dimno = 1, rpath%dimension
      before_gather(dimno) = rpath%rest_reference(1, dimno, repid)
    end do

#ifdef MPI
    call mpi_gather(before_gather, rpath%dimension, mpi_real8,&
                    after_gather,  rpath%dimension, mpi_real8,&
                    0, mpi_comm_airplane, ierror)

    if (main_rank) then
      do repid_i = 1, rpath%nreplica
        do dimno = 1, rpath%dimension
          path(repid_i,dimno) = after_gather((repid_i-1)*rpath%dimension+dimno)
        end do
      end do

      ! smooth
      path_smooth = path
      do repid_i = 2, rpath%nreplica - 1
        do dimno = 1, rpath%dimension
          path_smooth(repid_i,dimno) = (1.0_dp-rpath%smooth) * path(repid_i,dimno) + &
            (rpath%smooth*0.5_dp)*(path(repid_i-1,dimno) + path(repid_i+1,dimno))
        end do
      end do

      ! calc path length
      path_leng = 0.0_dp
      do repid_i = 2, rpath%nreplica
        path_leng(repid_i) = path_leng(repid_i-1) + &
          sqrt( sum( (path_smooth(repid_i,:)-path_smooth(repid_i-1,:))**2) )
        ! write(6,*) 'BEFORE inter_replica', &
        !             sqrt( sum( (path_smooth(repid_i,:)-path_smooth(repid_i-1,:))**2) )
      end do

      ! equi path length
      path_equi = 0.0_dp
      do repid_i = 2, rpath%nreplica
        path_equi(repid_i) = (repid_i-1)*path_leng(rpath%nreplica)/dble(rpath%nreplica -1)
      end do

      ! reparametrize
      path_reparm = path
      do repid_i = 2, (rpath%nreplica-1)
        do repid_j = 2, rpath%nreplica

          if ( (path_leng(repid_j-1) <  path_equi(repid_i)) .and. &
            (path_equi(repid_i))  <= path_leng(repid_j) ) then

            path_reparm(repid_i,:) = path_smooth(repid_j-1,:) + &
              (path_equi(repid_i) - path_leng(repid_j-1)) * &
              (path_smooth(repid_j,:) - path_smooth(repid_j-1,:)) / &
              sqrt( sum((path_smooth(repid_j,:) - path_smooth(repid_j-1,:))**2) )
            exit
          end if

        end do
      end do
    end if

    ! broadcast restraint reference
    call mpi_bcast(path_reparm, rpath%dimension*rpath%nreplica,&
                   mpi_real8, 0, mpi_comm_world, ierror)
#endif

    do dimno = 1, rpath%dimension
      rpath%rest_reference(1:2, dimno, repid) = path_reparm(repid, dimno)
    end do

    deallocate(path, path_smooth, path_reparm, path_leng, path_equi)

    return

  end subroutine reparametrize

end module at_rpath_mod
